import { compare } from "./Util.js";
import { float64, int32 } from "./Int32.js";
import { Option } from "./Option.js";
import { unfold, delay } from "./Seq.js";
import { op_Addition, fromZero } from "./BigInt.js";
import { BigInteger } from "./BigInt/z.js";
import { decimal, op_Addition as op_Addition_1, fromParts } from "./Decimal.js";
import { uint64, int64, op_Addition as op_Addition_2, fromBits } from "./Long.js";

export function makeRangeStepFunction<T>(step: T, stop: T, zero: T, add: (arg0: T, arg1: T) => T): (arg0: T) => Option<[T, T]> {
    const stepComparedWithZero: int32 = compare(step, zero) | 0;
    if (stepComparedWithZero === 0) {
        throw new Error("The step of a range cannot be zero");
    }
    const stepGreaterThanZero: boolean = stepComparedWithZero > 0;
    return (x: T): Option<[T, T]> => {
        const comparedWithLast: int32 = compare(x, stop) | 0;
        return ((stepGreaterThanZero && (comparedWithLast <= 0)) ? true : ((!stepGreaterThanZero) && (comparedWithLast >= 0))) ? [x, add(x, step)] : (void 0);
    };
}

export function integralRangeStep<T>(start: T, step: T, stop: T, zero: T, add: (arg0: T, arg1: T) => T): Iterable<T> {
    const stepFn: (arg0: T) => Option<[T, T]> = makeRangeStepFunction<T>(step, stop, zero, add);
    return delay((): Iterable<T> => unfold(stepFn, start));
}

export function rangeBigInt(start: BigInteger, step: BigInteger, stop: BigInteger): Iterable<BigInteger> {
    return integralRangeStep<BigInteger>(start, step, stop, fromZero(), op_Addition);
}

export function rangeDecimal(start: decimal, step: decimal, stop: decimal): Iterable<decimal> {
    return integralRangeStep<decimal>(start, step, stop, fromParts(0, 0, 0, false, 0), op_Addition_1);
}

export function rangeDouble(start: float64, step: float64, stop: float64): Iterable<float64> {
    return integralRangeStep<float64>(start, step, stop, 0, (x: float64, y: float64): float64 => (x + y));
}

export function rangeInt64(start: int64, step: int64, stop: int64): Iterable<int64> {
    return integralRangeStep<int64>(start, step, stop, fromBits(0, 0, false), op_Addition_2);
}

export function rangeUInt64(start: uint64, step: uint64, stop: uint64): Iterable<uint64> {
    return integralRangeStep<uint64>(start, step, stop, fromBits(0, 0, true), op_Addition_2);
}

export function rangeChar(start: string, stop: string): Iterable<string> {
    const intStop: int32 = stop.charCodeAt(0) | 0;
    const stepFn: (arg0: int32) => Option<[string, int32]> = (c: int32): Option<[string, int32]> => {
        if (c <= intStop) {
            return [String.fromCharCode(c), c + 1];
        }
        else {
            return void 0;
        }
    };
    return delay((): Iterable<string> => unfold(stepFn, start.charCodeAt(0)));
}

